{-# LANGUAGE BangPatterns #-}
{-# LANGUAGE LambdaCase #-}

module UIO.Plugin (plugin) where

import GHC.Plugins
import Data.Generics.Schemes (everywhereM)
import Data.Generics.Aliases (mkM)
import GHC.Types.Name
import Debug.Trace

isSpecificFunction :: Id -> Bool
isSpecificFunction f = case nameModule_maybe (idName f) of
  Just mod -> moduleNameString (moduleName mod) == "UIO" && getOccString f == "uniqueState"
  Nothing  -> False

plugin :: Plugin
plugin = defaultPlugin
    { installCoreToDos = install
    , pluginRecompile = purePlugin
    }

install :: [CommandLineOption] -> [CoreToDo] -> CoreM [CoreToDo]
install _ todos = do
  let myPass = CoreDoPluginPass ("Inline UIO.uniqueState") inlineSpecificFunction
  return $ case insertAfterLastCSE myPass todos of
    Nothing -> myPass : todos
    Just newTodos -> newTodos

insertAfterLastCSE :: CoreToDo -> [CoreToDo] -> Maybe [CoreToDo]
insertAfterLastCSE myPass = fmap reverse . go . reverse
  where go = \case
          [] -> Nothing
          passes@(CoreCSE {} : _) -> Just (myPass : passes)
          passes@(h@(CoreDoPasses subPasses) : t) -> case insertAfterLastCSE myPass subPasses of
            Nothing -> fmap (h :) $ insertAfterLastCSE myPass t
            Just newSubPasses -> Just $ CoreDoPasses newSubPasses : t
          h : t -> fmap (h :) $ insertAfterLastCSE myPass t

inlineSpecificFunction :: ModGuts -> CoreM ModGuts
inlineSpecificFunction guts = do
    let !binds = mg_binds guts
    inlinedBinds <- mapM (inlineSpecificFunctionBindM inlineSpecificFunctionTransform) binds
    return guts { mg_binds = inlinedBinds }

inlineSpecificFunctionTransform :: CoreExpr -> CoreM CoreExpr
inlineSpecificFunctionTransform e@(App (App (Var f) _) s)
  | isSpecificFunction f = return s
inlineSpecificFunctionTransform e = return e

inlineSpecificFunctionBindM :: (CoreExpr -> CoreM CoreExpr) -> CoreBind -> CoreM CoreBind
inlineSpecificFunctionBindM transform (NonRec b e) = do
    e' <- everywhereM (mkM transform) e
    return (NonRec b e')
inlineSpecificFunctionBindM transform (Rec pairs) = do
    pairs' <- mapM (\(b, e) -> everywhereM (mkM transform) e >>= \e' -> return (b, e')) pairs
    return (Rec pairs')
